#include "kiss.h"

/* Copyright (c) 2009, Zoe Stephenson.  All rights reserved, redistribution
 *  * permitted under the terms of the BSD license (see LICENSE for details) */

/*- Package description

 Kiss client handling system:

   The main kiss process will call the main client handler,

     int kiss_client_monitor(kiss_fifo client_input, kiss_config *config)

   as the last thing before exiting.

   When the client monitor exits, this is a trigger for kiss
   itself to exit, as the only thing which can legitiamtely
   ask kiss to close down is the client, on behalf of the
   user.

   The client monitor spawns the client named in config->kiss_client
   and attaches the fifo client_input to its standard input.
   If it fails to exec the client, it exits with status -1.

     The function calling the client monitor is responsible for
     creating the fifo.

   When the client terminates, the client monitor checks its
   return status.  
   
     A zero status causes the client monitor to
     exit with Success status, telling kiss to exit cleanly.

     Any other exit status causes the client monitor to
     attempt a respawn of the client.  This will be considered
     a client failure unless the client exited with a USR1 or 
     USR2 signal.  If this fails after RESPAWN_MAX tries, the 
     client monitor will exit with Failure status.

   The client monitor allocates its own fifo for IPC, and removes
   it when it exits.  The fifo tracker should find that this
   fifo is not present when it cleans up.


   When kiss wants to give information to the kiss client, it uses

     int signal_kiss_client(char *signaltext, FILE *clientstream)

   with the clientstream being a stream attached to the same fifo as
   client_input named earlier.  The trigger function is responsible for
   opening this stream.

   The signal text is expected to be a simple ASCII string with
   no delimiter.  A newline will be added to the signal text when
   it is written to the client's stdin stream.

*/


/*- Local configuration */

/* Number of times to respawn client process before giving up */
#define RESPAWN_MAX 5


/*- Types for local use */

/* ATC operation types */
typedef enum {KO_UNDEFINED=Undefined, KO_SET_STREAM, KO_CLOSE_STREAM, KO_SEND, KO_RECV} kiss_op;

/* ATC message types */
typedef enum {KV_UNDEFINED=Undefined, KV_EXIT, KV_EXIT_ERROR, KV_RESPAWN, KV_RESPAWN_ERROR} kiss_value;


/*- Prototypes for local use */

/* Signal handler to detect client exit (SIGCHLD) */
void client_signal_handler(int clientsignal);

/* Utility function for ATC from the signal handler to the monitor */
kiss_value kiss_client_operation(kiss_op op, kiss_fifo fifo_value, kiss_value value_value);


/*- Package contents */

/* Utility function for ATC from the signal handler to the monitor */
kiss_value kiss_client_operation(kiss_op op, kiss_fifo fifo_value, kiss_value value_value)
{
    static FILE
        *stream;
    kiss_value
        t;
    fd_set
        fds;
    int
        retval;

/* if op is KO_SET_STREAM, fifo_value is the name of the fifo to use
   for communication.  value_value is unused.  The return value is
   KV_UNDEFINED.

   if op is KO_CLOSE_STREAM, the fifo opened previously is closed.
   fifo_value and value_value are unused.  The return value is
   KV_UNDEFINED.

   if op is KO_SEND, the value in value_value is sent down the fifo.
   fifo_value is unused.  The return value is KV_UNDEFINED.

   if op is KO_RECV, the value sent down the fifo is returned.
   fifo_value and value_value are unused.
*/

    switch(op)
    {
    case KO_SET_STREAM:
        stream=fopen(fifo_value, "a+");
        return KV_UNDEFINED;
        break;
    case KO_CLOSE_STREAM:
        fclose(stream);
        return KV_UNDEFINED;
        break;
    case KO_SEND:
        /* Signal that the next kiss-client operation is according to
           value */
        (void)fwrite(&value_value, sizeof(kiss_value), 1, stream);
        fflush(stream);
        return KV_UNDEFINED;
        break;
    case KO_RECV:
        /* Grab kiss-client operation from stream */

        /* This only gets information from within the client's SIGCHLD
           handler.  As a result, whatever it is that is waiting for the
           data will be interrupted at the same time as the data becoming
           available.  To get around this, we have a select running
           beforehand to absorb the signal handler's interrupt, and then
           we inspect the data. */

        retval=0;
        /* Just cycle around and wait for the data to be available.
           We can only proceed if this happens, so we spin on it... */

        while(retval != 1)
        {
            FD_ZERO(&fds);
            FD_SET(fileno(stream), &fds);
            retval=select(fileno(stream)+1, &fds, NULL, NULL, NULL);
        }

        (void)fread(&t, sizeof(kiss_value), 1, stream);
        return t;
        break;
    default:
        error_warning(0, "Bad client monitor ATC operation");
        return KV_UNDEFINED;
        break;
    }
    return KV_UNDEFINED;
}

/* Signal handler to detect client exit (SIGCHLD) */
void client_signal_handler(int clientsignal)
{
    pid_t
        childpid;
    int
        childstatus;

/* When this handler catches a SIGCHLD, it checks the exit status
   of the client.  If the client exited normally, it sends KV_EXIT
   to the client monitor.  If the client exited with a SIGUSR1 or
   a SIGUSR2, it sends KV_RESPAWN to the client monitor.  If the
   client exited with a non-zero exit status, or exited due to any
   other signal, it sends KV_RESPAWN_ERROR to the client monitor.
*/
    switch(clientsignal)
    {
    case SIGCHLD:
        childpid=wait(&childstatus);
        /* We propagate the cause of exit from the client to the monitor */
        if (WIFEXITED(childstatus))
        {
            if (WEXITSTATUS(childstatus)==0)
            {
                /* exit kiss normal */
                kiss_client_operation(KO_SEND, NULLkiss_fifo, KV_EXIT);
            }
            else
            {
                /* client failure respawn */
                kiss_client_operation(KO_SEND, NULLkiss_fifo, KV_RESPAWN_ERROR);
            }
        }
        else if (WIFSIGNALED(childstatus))
        {
            switch (WTERMSIG(childstatus))
            {
            /*case SIGHUP: */ /* HUP got translated to an exit somehow... */
            case SIGUSR1:
            case SIGUSR2:
                /* client restart */
                kiss_client_operation(KO_SEND, NULLkiss_fifo, KV_RESPAWN);
                break;
            default:
                /* client failure respawn */
                kiss_client_operation(KO_SEND, NULLkiss_fifo, KV_RESPAWN_ERROR);
                break;
            }
        }
        else 
        {
            /* client failure respawn */
            kiss_client_operation(KO_SEND, NULLkiss_fifo, KV_RESPAWN_ERROR);
        }
        break;
    default:
        break;
    }
}

/* Kiss client monitor process - (re)spawn client and exit when client does */
response kiss_client_monitor(kiss_fifo client_input, kiss_config *config)
{
    pid_t
        clientpid;
    char
        *av[2];
    int
        respawns=0;
    kiss_fifo
        IPC_fifo;

    /* First, the IPC FIFO */
    IPC_fifo=new_fifo(config);
    kiss_client_operation(KO_SET_STREAM, IPC_fifo, KV_UNDEFINED);

    /* We loop until we have failed to create a working client
       process more than RESPAWN_MAX times.  If the process works
       properly, and exits successfully, we use that as a signal
       to exit kiss.
    */
    while(respawns < RESPAWN_MAX)
    {
        /* We have to reset the handler after each signal, for Linux */
        (void)signal(SIGCHLD, client_signal_handler);
        /* Spawn a client */
        if(0==(clientpid=fork()))
        {
            /* Child process is here */
            freopen(client_input, "r", stdin);
            av[0]=config->kiss_client;
            av[1]=NULLstring;
            execvp(config->kiss_client, av);
            /* If we reached here, we could not spawn the client.
               In this case, we exit Kiss abnormally... */
            kiss_client_operation(KO_SEND, NULLkiss_fifo, KV_EXIT_ERROR);
            error_fatal(0, "Client could not be executed in client process");
            /* Notreached */
        }
        /* Parent part - wait for the child to tell us something */
        switch(kiss_client_operation(KO_RECV, NULLkiss_fifo, KV_UNDEFINED))
        {
        case KV_EXIT:
            /* Exit cleanly please */
            kill_fifo(IPC_fifo);
            return SuccessResponse;
        case KV_EXIT_ERROR:
            /* Exit with an error */
            kill_fifo(IPC_fifo);
            return FailureResponse;
        case KV_RESPAWN:
            /* Respawn client without logging */
            break;
        case KV_RESPAWN_ERROR:
            /* Respawn client and log the fact */
            respawns++;
            break;
        default:
            error_fatal(0, "Bad client monitor message");
            break;
        }
    }
    /* All respawns exhausted */
    kill_fifo(IPC_fifo);
    return FailureResponse;
}

/* Write a string to the client process's standard input stream */
response signal_kiss_client(char *signaltext, FILE *s)
{
    if(0>=fprintf(s, "%s\n", signaltext))
    {
        error_warning(0, "Client stream unavailable");
        return FailureResponse;
    }
    if(0!=fflush(s))
    {
        error_warning(errno, "Client stream unavailable");
        return FailureResponse;
    }
    return SuccessResponse;
}
