#include "kiss.h"

/* Copyright (c) 2009, Zoe Stephenson.  All rights reserved, redistribution
 *  * permitted under the terms of the BSD license (see LICENSE for details) */

/*- Package description

 Kiss error reporting system:

   Whenever an error or a warning needs to be given, it will be
   produced using one of

     void error_fatal(int errornumber, char *description)
     void error_warning(int errornumber, char *description)

   A fatal error will cause kiss to exit, performing all necessary
   close-down actions.


   Warnings will be suppressed if indicated by setting the debug
   level appropriately.

*/

/*- Types for local use */

/* Error handler name operation types */
typedef enum {EH_UNDEFINED=Undefined, EH_SET, EH_GET, EH_MAX} eh_op;

/*- Prototypes for local use */

/* Manipulate a stored string representing the current program name */
char *error_pname(eh_op op, char *string);

/* Perform all actions associated with a clean exit from the program */
void error_actions(void);


/*- Package contents */

/* Initialise the error system and supply the current program name */
void error_init(char *programname)
{
    (void)error_pname(EH_SET, programname);
}

/* Spew out a fatal error and close */
void error_fatal(int errornumber, char *description)
{
    if (errornumber != 0)
    {
        fprintf(stderr, "%s: Fatal: %s (%s)\n", error_pname(EH_GET, NULLstring), description, strerror(errornumber));
    }
    else
    {
        fprintf(stderr, "%s: Fatal: %s\n", error_pname(EH_GET, NULLstring), description);
    }
    error_actions();
    exit(EXIT_FAILURE);
}

/* Spew out a warning, but carry on */
void error_warning(int errornumber, char *description)
{
    if (errornumber != 0)
    {
        fprintf(stderr, "%s: Warning: %s (%s)\n", error_pname(EH_GET, NULLstring), description, strerror(errornumber));
    }
    else
    {
        fprintf(stderr, "%s: Warning: %s\n", error_pname(EH_GET, NULLstring), description);
    }
}

/* Manipulate a stored string representing the current program name */
char *error_pname(eh_op op, char *string)
{
    static char
        *pn=NULLstring;
    
    switch(op)
    {
    case EH_SET:
        if (pn!=NULLstring)
        {
            free((any)pn);
        }
        pn=(char *)malloc(strlen(string)+1);
        strcpy(pn, string);
        break;
    case EH_GET:
        break;
    default:
        fprintf(stderr, "Error: trouble in the error handler.\n");
        exit(EXIT_FAILURE);
        break;
    }
    return pn;
}

/* Perform all actions associated with a clean exit from the program */
void error_actions(void)
{
    static int
        recursion_flag=0;
    
    /* We protect against the possibility that errors may occur while
       we're cleaning up (if things have gone horribly wrong! */
    if(0==recursion_flag)
    {
        recursion_flag=1;
        /* Now try to do a clean shutdown */
        clean_all_fifos();
        recursion_flag=0;
    }
    else
    {
        fprintf(stderr, "Error: Recursion in the error handler.\n");
    }
}

