#include "kiss.h"

/* Copyright (c) 2009, Zoe Stephenson.  All rights reserved, redistribution
 *  * permitted under the terms of the BSD license (see LICENSE for details) */

/*- Package description

 Kiss fifo ADT:

   Whenever the kiss system needs a fifo to communicate between
   its separate parts, it should call

     kiss_fifo new_fifo(kiss_config *config)

   This will allocate a new unique fifo name not currently in use
   by this process, and create that fifo in the filesystem.

   Multiple instances of kiss running as the same user on the
   same display will have fifo names which clash, and this is
   undesirable.  Do not run kiss more than once per display.


   When a fifo is no longer required, 

     void kill_fifo(kiss_fifo f)

   should be used to remove it from the filesystem.


   When kiss is exiting, it should call

     void clean_all_fifos(void)

   to ensure that no stale fifos are left in the filesystem from
   this process.

*/


/*- Local configuration */

/* Number of characters to expect in a pathname */
#define PATHNAME_MAX 1024

/* Stem for fifo names in the FS */
#define FIFO_NAME_STEM "/kiss_fifo"

/* Maximum failed FIFOs we're willing to allow */
#define MAX_UNDEFINED_FIFOS 5


/*- Types for local use */

typedef enum {IU_UNDEFINED=Undefined, IU_QUERY, IU_SET, IU_CLEAR} inuse_op;

struct name_usage;
struct name_usage
{
    char 
        *inuse_name;
    struct name_usage
        *next;
};
typedef struct name_usage name_usage;
#define NULLname_usage ((name_usage *)0)
name_usage *new_name_usage(void);


/*- Prototypes for local use */

/* Remove all fifos created by kiss from the filesystem */
void clean_all_fifos(void);

/* ADT to keep track of the fifos in use by this kiss process */
Bool inuse(inuse_op op, char *name);

/* Create a fifo name based on an integer, and report on its viability */
Bool makefifoname(int which, char *buffer, kiss_config *config);

/*- Package contents */

/* ADT to keep track of the fifos in use by this kiss process */
Bool inuse(inuse_op op, char *name)
{
    static name_usage
        *table=NULLname_usage;
    
    switch(op)
    {
    case IU_CLEAR:
        PROTECT(table)
        {
            name_usage
                *r,
                *q=table;
            
            while(q!=NULLname_usage)
            {
                r=q->next;
                /* We're not concerned about whether it goes - we just try */
                (void)unlink(q->inuse_name);
                free((any)q->inuse_name);
                free((any)q);
                q=r;
            }
            table=NULLname_usage;
        }
        return Undefined;
        break;
    case IU_SET:
        PROTECT(table)
        {
            char
                *s;
            name_usage
                *q=table,
                *p=new_name_usage();

            PROTECT(p)
            {
                /* Fill in */
                p->next=NULLname_usage;
                s=(char *)malloc(strlen(name)+1);
                strcpy(s, name);
                p->inuse_name=s;
                /* Link to the end */
                while(q->next!=NULLname_usage)
                {
                    q=q->next;
                }
                q->next=p;
                return True;
            }
            else
            {
                error_warning(0, "Fifo name usage allocation failed");
                return Undefined;
            }
        }
        else
        {
            table=new_name_usage();
            PROTECT(table)
            {
                name_usage
                    *p=new_name_usage();
                PROTECT(p)
                {
                    /* Fill in direct */
                    table->next=NULLname_usage;
                    table->inuse_name=(char *)malloc(strlen(name)+1);
                    strcpy(table->inuse_name, name);
                    return True;
                }
                else
                {
                    error_warning(0, "Fifo name usage allocation failed");
                    return Undefined;
                }
            }
            else
            {
                error_warning(0, "Fifo name usage allocation failed");
                return Undefined;
            }
        }
        break;
    case IU_QUERY:
        PROTECT(table)
        {
            name_usage
                *p=table;
            
            while(p!=NULLname_usage)
            {
                if(strcmp(name, p->inuse_name)==0)
                {
                    return True;
                }
                p=p->next;
            }
            return False;
        }
        else
        {
            /* Quite OK if we ask for a FIFO name before the
               table's been populated. */
            return False;
        }
        break;
    default:
        error_warning(0, "Fifo name usage operation broken");
        return Undefined;
        break;
    }
}

/* Create a fifo name based on an integer, and report on its viability */
Bool makefifoname(int which, char *buffer, kiss_config *config)
{
    /* We'll return True if we could use this name, False if we
       can't use this name, and Undefined if we could use the name,
       but couldn't create the FIFO. */
    static char
        fifoname[]=FIFO_NAME_STEM;
    char
        number[10];
    struct stat
        statbuf;

    if (config->kiss_fifos == NULLstring)
    {
        return Undefined;
    }
    strcpy(buffer, config->kiss_fifos);
    strcat(buffer, fifoname);
    strcat(buffer, config->kiss_display);
    sprintf(number, "-%d", getuid());
    strcat(buffer, number);
    sprintf(number, "-%d", which);
    strcat(buffer, number);


    /* We now have a name.  Is this name being used by this process,
       or is it somebody else's? */

    if (True==inuse(IU_QUERY, buffer))
    {
        /* We already have this filename, return an 'error' */
        return False;
    }
    if (0==stat(buffer, &statbuf))
    {
        /* This fifo exists (or a file exists, anyway).  We should
           remove it if possible.  This is a toss-up between having
           two identical kiss processes garble at each other, and
           leaving files around in the filesystem. */
        if(0!=unlink(buffer))
        {
            /* It couldn't be removed.  We can't use this name */
            return False;
        }
    }
    /* Now try to create the FIFO */
    if(0==mknod(buffer, (0600 | S_IFIFO), (dev_t)0))
    {
        /* Successful create, log the name and return 0 */
        (void)inuse(IU_SET, buffer);
        return True;
    }
    else
    {
        error_warning(0, "Viable FIFO couldn't be created");
        return Undefined;
    }
}

/* Allocate a new fifo name in the filesystem */
kiss_fifo new_fifo(kiss_config *config)
{
    kiss_fifo
        f;
   /* I'm sure there's a system define for the length of a pathname... */ 
    int
        r,
        ucount=0,
        k=0;
    char
        fifo_name_buffer[PATHNAME_MAX];

    /* Generate a pathname */

    r=False;
    while(r!=True && ucount <= MAX_UNDEFINED_FIFOS)
    {
        r=makefifoname(k, fifo_name_buffer, config);
        if (r==Undefined)
        {
            ucount++;
        }
        k++;
    }
    if (r!=True)
    {
        error_fatal(0, "Fifo allocation failed");
    }

    /* Make a copy of the name to return */
    f=(kiss_fifo)malloc(strlen(fifo_name_buffer)+1);
    strcpy(f, fifo_name_buffer);
    return f;
}

/* Remove a FIFO from the filesystem and free its name */
void kill_fifo(kiss_fifo f)
{
    (void)unlink(f);
    free((any)f);
}

/* Remove all fifos created by kiss from the filesystem */
void clean_all_fifos(void)
{
    (void)inuse(IU_CLEAR, NULLstring);
}

/* Local ADT methods */

name_usage *new_name_usage(void)
{
    name_usage
        *k=NEW(name_usage);
    PROTECT(k)
    {
        k->inuse_name=NULLstring;
        k->next=NULLname_usage;
    }
    return k;
}
