#include "kiss.h"

/* Copyright (c) 2009, Zoe Stephenson.  All rights reserved, redistribution
 *  * permitted under the terms of the BSD license (see LICENSE for details) */

/*- Main Body

  This main process performs the following:

    Parse parameters
    Open X display connection
    Allocate fifos
    Create monitor and trigger processes
    Spawn a client and wait for it to finish
    Close down

  The processes are:

    1 Monitor: Process to monitor the root window and detect kiss events;
    2 Trigger: Process to combine multiple kiss events into kiss sequences;
    3  Client: Process to act upon kiss sequences.

   The client process, process 3, is actually a monitor wrapper around
   a further client - it manages the cases where the client dies due
   to unforeseen circumstances.  This client monitor is responsible for 
   the control of kiss - when it exits, kiss also exits. 
*/

int main(int argc, char *argv[])
{
    response
        exitstatus;
    pid_t
        monitorchild,
        triggerchild;

    kiss_fifo
        monitor_to_trigger,
        trigger_to_client;

    kiss_parameter_list
        *parameters=new_kiss_parameter_list();
    kiss_config
        *config=new_kiss_config();
    kiss_target
        *target=new_kiss_target();

    /* Initialise error-handling strategies */

    error_init(argv[0]);

    /* Create configuration */

    add_kiss_parameter(parameters, DisplayParameter, "DISPLAY", "-display", DEFAULT_KISS_DISPLAY);
    add_kiss_parameter(parameters, ScreenParameter, "KISS_SCREEN", "-screen", DEFAULT_KISS_SCREEN);
    add_kiss_parameter(parameters, DelayParameter, "KISS_DELAY", "-delay", DEFAULT_KISS_DELAY);
    add_kiss_parameter(parameters, ClientParameter, "KISS_CLIENT", "-client", DEFAULT_KISS_CLIENT);
    add_kiss_parameter(parameters, FifoParameter, "KISS_FIFOS", "-fifos", DEFAULT_KISS_FIFOS);

    parse_parameters(parameters, config, argc, argv);
    
    /* Open display connection */

	if ((target->kiss_Xdisplay=XOpenDisplay(config->kiss_display))==NULL)
	{
        
        error_fatal(0, "Cannot open X display");
	}

    /* Propagate the kiss display to client environments */
    /* TODO: Check to see whether this actually works */
    /* TODO: Ensure that the screen number gets propagated as well */

    setenv("DISPLAY", config->kiss_display, 0);

	/* Check for a bad screen specifier and switch to display's default 
       screen instead.  We get a bad screen if the screen argument was
       parsed badly, but we couldn't do this back then, as the display
       information hadn't yet been parsed. */

    if (config->kiss_screen == KISS_BAD_SCREEN)
    {
        target->kiss_Xscreen=DefaultScreen(target->kiss_Xdisplay);
    }
    else
    {
        target->kiss_Xscreen=config->kiss_screen;
    }

    /* Further sanity check on the screen number */

    if ((target->kiss_Xscreen < 0) || (target->kiss_Xscreen >= ScreenCount(target->kiss_Xdisplay)))
    {
        error_fatal(0, "Cannot access screen");
    }

    /* Find the root window */

    target->kiss_RootWindow=RootWindow(target->kiss_Xdisplay, target->kiss_Xscreen);

    /* Get size of the root window */

    target->minX=0;
    target->minY=0;
    target->maxX=DisplayWidth(target->kiss_Xdisplay, target->kiss_Xscreen)-1;
    target->maxY=DisplayHeight(target->kiss_Xdisplay, target->kiss_Xscreen)-1;

    /* Initialise IPC fifos */

    monitor_to_trigger=new_fifo(config);
    trigger_to_client=new_fifo(config);

    /* Perform processing */

    /* Monitor the X events on the root window for kiss events */
    switch(monitorchild=fork())
    {
    case -1:
        error_fatal(0, "Fork for monitor process");
        break;
    case 0:
        kiss_monitor(monitor_to_trigger, target);
        /* Shouldn't exit */
        exit(EXIT_FAILURE);
        break;
    }

    /* Convert sequences of kiss events into a kiss sequence */
    switch(triggerchild=fork())
    {
    case -1:
        error_fatal(0, "Fork for trigger process");
        break;
    case 0:
        kiss_trigger(monitor_to_trigger, trigger_to_client, config, target);
        /* Shouldn't exit */
        exit(EXIT_FAILURE);
    }

    /* Run the client and wait for it to exit */
    exitstatus=kiss_client_monitor(trigger_to_client, config);
    
    /* Remove child processes, as they're no longer needed */
    kill(monitorchild, SIGTERM);
    kill(triggerchild, SIGTERM);

    /* We can now close the display, too */
    (void)XCloseDisplay(target->kiss_Xdisplay);

    clean_all_fifos();

    switch(exitstatus)
    {
    case SuccessResponse:
        exit(EXIT_SUCCESS);
        /* Notreached */
        break;
    case FailureResponse:
        exit(EXIT_FAILURE);
        /* Notreached */
        break;
    default:
        exit(EXIT_FAILURE);
        /* Notreached */
        break;
    }
}

