#include "kiss.h"

/* Copyright (c) 2009, Zoe Stephenson.  All rights reserved, redistribution
 *  * permitted under the terms of the BSD license (see LICENSE for details) */

/*- Package description

 Kiss event monitor:

   The main kiss process will run in parallel with

     void kiss_monitor(kiss_fifo output, kiss_target *target)

   when it starts up.  It is responsible for creating a
   separate thread of control.

   The kiss event monitor will wait for the mouse pointer
   on the target to reach the edge of the screen, and announce
   it on the output fifo.

     The kiss event monitor will create a stream for the
     output fifo.

   This function is not expected to exit.  If it does, the main
   kiss process is responsible for detecting this as an error
   condition.

   The behaviour of this function defines the semantics by which
   mouse movements are interpreted as kiss events.

*/


/*- Prototypes for local use */

/* Determine whether the pointer is at a particular side of the screen */
Bool evaluateSide(side whichside, kiss_target *target, XPointerMovedEvent *current_event);


/*- Package contents */

/* Monitor the target for kiss events and annouce them */
void kiss_monitor(kiss_fifo output, kiss_target *target)
{
    FILE
        *outputstream;
    XEvent
        eventBuffer;
    kiss_state
        *current_state=new_kiss_state(),
        *previous_state=new_kiss_state();

    outputstream=fopen(output, "w");
    if (outputstream==(FILE *)NULL)
    {
        error_fatal(errno, "Monitor output stream unavailable");
    }


    XSelectInput(target->kiss_Xdisplay, target->kiss_RootWindow, PointerMotionMask);

	for(;;)
	{
		XNextEvent(target->kiss_Xdisplay, &eventBuffer);
		if (eventBuffer.type==MotionNotify)
        {
            side
                this_side,
                s;
            kiss_state
                *temp;
            XPointerMovedEvent
                *current_event;

            current_event=(XPointerMovedEvent *) &eventBuffer;

            /* Reuse old previous state */
            temp=current_state;
            current_state=previous_state;
            previous_state=temp;

            /* Check where the event occured */
            for (s=NoSide; s<MaxSide; s++)
            {
                if (s>NoSide)
                {
                    current_state->location[s]=evaluateSide(s, target, current_event);
                }
            }

            /* Now check to see if that was a trigger */
            this_side=NoSide;
            for (s=NoSide; s<MaxSide; s++)
            {
                if (s>NoSide)
                {
                    if ((True==current_state->location[s]) && (True!=previous_state->location[s]))
                    {
                        /* This is a trigger on side s */
                        this_side=s;
                        break;
                    }
                }
            }
            /* And if is was, send the trigger */
            if (this_side != NoSide)
            {
                kiss
                    mykiss;
                
                mykiss.x=current_event->x_root;
                mykiss.y=current_event->y_root;
                mykiss.time=current_event->time;
                mykiss.state=current_event->state;

                /* We've got a kiss trigger on side <this_side> */
                fwrite(&mykiss, sizeof(kiss), 1, outputstream);
                fflush(outputstream);
            }
        }
    }
}

/* Determine whether the pointer is at a particular side of the screen */
Bool evaluateSide(side whichside, kiss_target *target, XPointerMovedEvent *current_event)
{
    switch(whichside)
    {
    case LeftSide:
        return MakeBool(current_event->x_root == target->minX);
        break;
    case RightSide:
        return MakeBool(current_event->x_root == target->maxX);
        break;
    case TopSide:
        return MakeBool(current_event->y_root == target->minY);
        break;
    case BottomSide:
        return MakeBool(current_event->y_root == target->maxY);
        break;
    default:
        error_warning(0, "Side evaluation broken");
        return Undefined;
        break;
    }
}
