#include "kiss.h"

/* Copyright (c) 2009, Zoe Stephenson.  All rights reserved, redistribution
 *  * permitted under the terms of the BSD license (see LICENSE for details) */

/*- Package description

 Kiss parameters:

   The kiss process will announce the names of parameters it expects
   to use with 

     void add_kiss_parameter(kiss_parameter_list *params, parameter_id id, char *eName, char *aName, char *dValue)

   The parameter information will be added to the queue params.

   Each parameter is identified to the configuration system by a
   parameter id.

   The parameters eName, aName and dValue are the environment name,
   argument name and default value, respectively.


   Once the parameters are announced, the main kiss process will
   analyse the environment and the command-line with

     void parseParameters(kiss_parameter_list *params, kiss_config *config, int argc, char *argv[]) 

   The configuration at config will be created according to the
   parameters listed in params, taking information from argv.

   This function checks for a kiss client in the user's home
   directory and uses that as the client if it exists.

*/

/*- Local configuration */

/* Maximum length of a commandline */
#define CMDLINE_MAX 1024

/* Minimum feasible delay for human input */
#define MIN_DELAY 200

/*- Prototypes for local use */

/* Allocate a parameter structure */
kiss_parameter *MakeParameter(parameter_id id, char *eName, char *aName, char *dValue);

/* Find the value of a single parameter */
void parseOneParameter(kiss_parameter *p, int argc, char *argv[]);


/*- Package contents */

/* Add a new parameter to a list of parameters */
void add_kiss_parameter(kiss_parameter_list *params, parameter_id id, char *eName, char *aName, char *dValue)
{
    kiss_parameter_list
        *p;
    
    if (params->content==NULLkiss_parameter)
    {
        /* The queue is an empty record, so fill that one first... */
        params->content=MakeParameter(id, eName, aName, dValue);
    }
    else
    {
        kiss_parameter_list
            *q_entry=NEW(kiss_parameter_list);

        PROTECT(q_entry)
        {
            q_entry->next=NULLkiss_parameter_list;
            q_entry->content=MakeParameter(id, eName, aName, dValue);
            /* Find the last entry in the queue */
            p=params;
            PROTECT(p)
            {
                while (p->next != NULLkiss_parameter_list)
                {
                    p=p->next;
                }
                
                /* Make it point at the new one */
                p->next=q_entry;
            }
            else
            {
                error_warning(0, "Parameter list upset");
            }
        }
    }
}

/* Allocate a parameter structure */
kiss_parameter *MakeParameter(parameter_id id, char *eName, char *aName, char *dValue)
{
    kiss_parameter
        *param=new_kiss_parameter();
    
    PROTECT(param)
    {
        param->environmentName=malloc(strlen(eName)+1);
        strcpy(param->environmentName, eName);
        param->argumentName=malloc(strlen(aName)+1);
        strcpy(param->argumentName, aName);
        param->defaultValue=malloc(strlen(dValue)+1);
        strcpy(param->defaultValue, dValue);
        param->currentValue=(char *)0;
        param->id=id;
    }
    else
    {
        error_warning(0, "Parameter allocation failed");
    }
    return param;
}

/* Create a configuration based on parameters */
void parse_parameters(kiss_parameter_list *params, kiss_config *config, int argc, char *argv[]) 
{
    kiss_parameter_list
        *current=params;
    kiss_parameter
        *thisParam;
    int 
        param; 

    /* Obtain parameter values */
    
    while(current != NULLkiss_parameter_list)
    {
        parseOneParameter(current->content, argc, argv);
        current=current->next;
    }

    /* Now turn them into a config-structure */

    current=params;
    while(current != NULLkiss_parameter_list)
    {
        thisParam=current->content;
        PROTECT(thisParam)
        {
            switch(thisParam->id)
            {
            case DisplayParameter:
                config->kiss_display=thisParam->currentValue;

                break;
            case ScreenParameter:
                if (sscanf(thisParam->currentValue, "%i", &param)>0)
                {
                    config->kiss_screen=param;
                }
                break;
            case DelayParameter:
                if (sscanf(thisParam->currentValue, "%i", &param)>0)
                {
                    config->kiss_delay=param;
                }
                break;
            case ClientParameter:
                config->kiss_client=thisParam->currentValue;
                if (thisParam->currentValue==thisParam->defaultValue)
                {
                    /* This is still at default, is there a user-specific
                       override present? */
                    char
                        *homedir;

                    if ((homedir=getenv("HOME")) != NULLstring)
                    {
                        char
                            *userclient;
                        struct stat
                            buffer;

                        userclient=(char *)malloc(FILENAME_MAX);
                        strncpy(userclient, homedir, FILENAME_MAX);
                        strncat(userclient, "/" DEFAULT_KISS_USER_CLIENT, FILENAME_MAX);

                        /* Does that exist?  Is it executable and readable? */
                        if ((0==stat(userclient, &buffer)) && (buffer.st_mode & (S_IRUSR | S_IXUSR)))
                        {
                            config->kiss_client = userclient;
                        }
                        else
                        {
                            free((any)userclient);
                        }
                    }
                    else
                    {
                        error_warning(0, "Could not locate your home directory");
                    }
                }
                break;
            case FifoParameter:
                /* Check to make sure we could put fifos here */
                {
                    struct stat
                        statbuf;
                    
                    /* Here, I'm not going to worry about who can write to this
                       directory.  I'll pick that up later when I try and put
                       things there.  But, I don't want to try if it's not
                       even a directory */
                    if (0==stat(thisParam->currentValue, &statbuf) && S_ISDIR(statbuf.st_mode))
                    {
                        config->kiss_fifos=thisParam->currentValue;
                    }
                    else
                    {
                        error_warning(0, "Fifo parameter is not a directory");
                    }
                }
                break;
            default:
                error_warning(0, "Bad parameter type");
                break;
            }
        }
        current=current->next;
    }

    /* And then check that structure for sanity */
    if (config->kiss_delay <= MIN_DELAY)
    {
        error_warning(0, "Delay parameter infeasibly small");
        config->kiss_delay=MIN_DELAY;
    }
}

/* Find the value of a single parameter */
void parseOneParameter(kiss_parameter *p, int argc, char *argv[])
{
    int
        arg;
    char
        *temp;

    /* First, check the arguments for argument p->argumentName */

    for (arg=1; arg<argc; arg++)
	{
        /* Check for -<word> <argument> */
		if ((0==strncmp(argv[arg], p->argumentName, CMDLINE_MAX)) && (arg<argc-1))
		{
			p->currentValue=(char *)malloc(strlen(argv[arg+1])+1);
			strcpy(p->currentValue, argv[arg+1]);
            return;
        }

        /* Check for --<word>=<argument> */
        temp=argv[arg];
        if (*temp++=='-')
        {
            if (*temp=='-')
            {
                int
                    l=strlen(p->argumentName);
                

                if (0==strncmp(temp, p->argumentName, l))
                {
                    char
                        *q=temp+l;
                    
                    if (*q++=='=')
                    {
                        p->currentValue=(char *)malloc(strlen(q)+1);
                        strcpy(p->currentValue, q);
                        return;
                    }
                }
            }
        }
    }

    /* Now, check the environment for p->environmentName */

    if ((temp=getenv(p->environmentName)) != NULLstring)
    {
        p->currentValue=(char *)malloc(strlen(temp)+1);
		strcpy(p->currentValue, temp);
        return;
	}
    
    /* Otherwise, point currentValue to defaultValue */

    p->currentValue=p->defaultValue;
    return;
}
