#include "kiss.h"

/* Copyright (c) 2009, Zoe Stephenson.  All rights reserved, redistribution
 *  * permitted under the terms of the BSD license (see LICENSE for details) */

/*- Package description

 Kiss trigger generator:

   The main kiss process will run in parallel with

     void kiss_trigger(kiss_fifo input, kiss_fifo output, kiss_config *config)
   
   when it starts up.  It is responsible for creating a
   separate thread of control.

   The kiss trigger generator will wait for kiss events to
   be announced on the input fifo.  When a sequence of such events,
   separated by delays of no more than config->kiss_delay, has
   been detected, it will create and send a string to the client
   on the output fifo, representing this sequence of kiss events.

     The kiss trigger generator will create a stream for each
     of the input and output fifos.

   This function is not expected to exit.  If it does, the main
   kiss process is responsible for detecting this as an error
   condition.

   The behaviour of this function defines the semantics by which
   a sequence of kiss events is interpreted as a single kiss trigger.

*/


/*- Local configuration */

/* Characters in the string representation of a single kiss event */
#define ONE_KISS_MAX 100


/*- Prototypes for local use */

/* Convert a kiss queue to a string and write it to the output stream */
void announce_kiss(kiss_list *queue, FILE *outputstream, kiss_target *target);

/* Convert a kiss's content to a letter representing the side */
char *str_kiss_side(kiss *k, kiss_target *target);

/* Convert the keyboard state in a kiss to a string of [CLS12345] */
char *str_kiss_kstate(kiss *k);

/* Convert the mouse state in a kiss to a string of [12345] */
char *str_kiss_mstate(kiss *k);

/* Calculate the difference between two times */
int kiss_time_sub(Time a, Time b);


/*- Package contents */

/* Kiss trigger generator - detect and announce kiss event sequences */
void kiss_trigger(kiss_fifo input, kiss_fifo output, kiss_config *config, kiss_target *target)
{
    FILE
        *inputstream,
        *outputstream;
    kiss_list
        *tempq,
        *q,
        *queue=NULLkiss_list;
    int
        retval;
    fd_set
        fds;
    struct timeval
        to;

    /* Open streams */
    inputstream=fopen(input, "r");
    if (inputstream == (FILE *)NULL)
    {
        error_fatal(errno, "Trigger input stream unavailable");
    }
    outputstream=fopen(output, "w");
    if (outputstream == (FILE *)NULL)
    {
        error_fatal(errno, "Trigger output stream unavailable");
    }

    for(;;)
    {
        /* Clear the queue for the next kiss */
        q=queue;
        while(q!=NULLkiss_list)
        {
            tempq=q->next;
            free((any)q);
            q=tempq;
        }
        queue=new_kiss_list();
        queue->content=new_kiss();

        /* Wait for a kiss to come in on the input fifo */
                
        FD_ZERO(&fds);
        FD_SET(fileno(inputstream), &fds);
        retval=select(fileno(inputstream)+1, &fds, NULL, NULL, NULL);

        fread(queue->content, sizeof(kiss), 1, inputstream);

        if ((0!=ferror(inputstream)) || (0!=feof(inputstream)))
        {
            error_fatal(errno, "Trigger input stream error");
        }
        else
        {
            /* Now wait for a little bit longer, and see if there's
               any more kisses coming in... */
            kiss_list
                *q=queue;
            int
                further=True;

            while(further)
            {
                FD_ZERO(&fds);
                FD_SET(fileno(inputstream), &fds);
                to.tv_sec=config->kiss_delay/1000; 
                to.tv_usec=1000*(config->kiss_delay%1000);
                retval=select(fileno(inputstream)+1, &fds, NULL, NULL, &to);
                if (retval>0)
                {
                    /* Another kiss occured within the delay - store and repeat */
                    q->next=new_kiss_list();
                    q=q->next;
                    q->next=NULLkiss_list;
                    q->content=new_kiss();

                    fread(q->content, sizeof(kiss), 1, inputstream);
                    if ((0!=ferror(inputstream)) || (0!=feof(inputstream)))
                    {
                        error_fatal(errno, "Trigger input stream error");
                    }
                }
                else
                {
                    /* Nothing Interesting happened - exit the loop and trigger */
                    further=False;
                }
            }
            /* Now check the kiss cache, construct a string and trigger */

            announce_kiss(queue, outputstream, target);
        
        }
    }
}

/* Convert a kiss queue to a string and write it to the output stream */
void announce_kiss(kiss_list *queue, FILE *outputstream, kiss_target *target)
{
/* The kiss queue info is as follows:

    <number of entries> (Per-entry:){
      <side> <x> <y> <kstate> <mstate> <time since previous (0 for first)>
    }

    number of entries: decimal
    side: [TBLR]
    x: decimal
    y: decimal
    kstate: [CLS12345]*
    mstate: [12345]*
    time: decimal

*/
    int
        first=True,
        number=0;
    char
        space[]=" ",
        singlebuffer[ONE_KISS_MAX],
        *eventstring,
        *kq;
    Time
        lasttime;
    kiss_list
        *q=queue;
    
    kq=(char *)malloc(2); /* Start off with a little buffer */
    *kq=0;                /* With a null string in it */

    PROTECT(queue)
    {
        PROTECT(queue->content)
        {
            lasttime=queue->content->time;  /* Initial last-time is first event */
        }

        for(; q!=NULLkiss_list; q=q->next)
        {
            PROTECT(q->content)
            {
                sprintf(singlebuffer, "%s %d %d %s %s %d", str_kiss_side(q->content, target), q->content->x, q->content->y, str_kiss_kstate(q->content), str_kiss_mstate(q->content), kiss_time_sub(q->content->time, lasttime));
                kq=(char *)realloc((any)kq, strlen(kq)+strlen(singlebuffer)+2);
                if (first)
                {
                    first=False;
                }
                else
                {
                    strcat(kq, space);
                }
                strcat(kq, singlebuffer);

                lasttime=q->content->time;
                number++;
            }
        }
        /* Now, build the final string */
        /* TODO: remove this ugly constant */
        eventstring=malloc(strlen(kq)+10);
        sprintf(eventstring, "%d ", number);
        strcat(eventstring, kq);

        /* Announce it */
        (void)signal_kiss_client(eventstring, outputstream);
        free((any)eventstring);
    }
    /* And free the string */

    free((any)kq);

}

/* Convert a kiss's content to a letter representing the side */
char *str_kiss_side(kiss *k, kiss_target *target)
{
    static char
        buffer[2];
    
    buffer[0]='-';
    buffer[1]=(char)0;
    PROTECT(k)
    {
        PROTECT(target)
        {
            if (k->x == target->minX)
            {
                sprintf(buffer, "L");
            }
            else if (k->x == target->maxX)
            {
                sprintf(buffer, "R");
            }
            else if (k->y == target->minY)
            {
                sprintf(buffer, "T");
            }
            else if (k->y == target->maxY)
            {
                sprintf(buffer, "B");
            }
            else
            {
                sprintf(buffer, "?");
            }
        }
    }
    return buffer;
}

/* Convert the keyboard state in a kiss to a string of [CLS12345] */
char *str_kiss_kstate(kiss *k)
{
    static char
        buffer[10];
    
    buffer[0]=(char)0;

    PROTECT(k)
    {
        if (k->state & ControlMask)
        {
            strcat(buffer, "C");
        }
        if (k->state & LockMask)
        {
            strcat(buffer, "L");
        }
        if (k->state & ShiftMask)
        {
            strcat(buffer, "S");
        }
        if (k->state & Mod1Mask)
        {
            strcat(buffer, "1");
        }
        if (k->state & Mod2Mask)
        {
            strcat(buffer, "2");
        }
        if (k->state & Mod3Mask)
        {
            strcat(buffer, "3");
        }
        if (k->state & Mod4Mask)
        {
            strcat(buffer, "4");
        }
        if (k->state & Mod5Mask)
        {
            strcat(buffer, "5");
        }
    }
    if(buffer[0]==(char)0)
    {
        buffer[0]='-';
        buffer[1]=(char)0;
    }
    return buffer;
}

/* Convert the mouse state in a kiss to a string of [12345] */
char *str_kiss_mstate(kiss *k)
{
    static char
        buffer[7];
    
    buffer[0]=(char)0;
    PROTECT(k)
    {
        if (k->state & Button1Mask)
        {
            strcat(buffer, "1");
        }
        if (k->state & Button2Mask)
        {
            strcat(buffer, "2");
        }
        if (k->state & Button3Mask)
        {
            strcat(buffer, "3");
        }
        if (k->state & Button4Mask)
        {
            strcat(buffer, "4");
        }
        if (k->state & Button5Mask)
        {
            strcat(buffer, "5");
        }
    }
    if(buffer[0]==(char)0)
    {
        buffer[0]='-';
        buffer[1]=(char)0;
    }
    return buffer;
}

/* Calculate the difference between two times */
int kiss_time_sub(Time a, Time b)
{
    /* Here, a is supposed to be the more recent */
    if (a==b)
    {
        return 0;
    }
    if (a>b)
    {
        return a-b;
    }
    else
    {
        /* Server time is unsinged long, and that could wrap, so we
           rely on a bit of overflow here... probably not too portable,
           this, but then time that wraps round is stupid anyway. */
        return (int)(a+(long)(b));
    }
}

